@@ -7,6 +7,7 @@ import time
 from moto import mock_s3
 
 def rmtree(path):
+  path = path.replace("file://", "")
   if os.path.exists(path):
     shutil.rmtree(path)
 
@@ -40,12 +41,7 @@ def s3(aws_credentials):
 @pytest.mark.parametrize("protocol", ('mem', 'file', 's3'))#'gs'))
 def test_read_write(s3, protocol, num_threads, green):
   from cloudfiles import CloudFiles, exceptions
-  if protocol == 'file':
-    path = '/tmp/cloudfiles/rw'
-    rmtree(path)
-    url = 'file://' + path
-  else:
-    url = '{}://cloudfiles/rw'.format(protocol)
+  url = compute_url(protocol, "rw")
 
   cf = CloudFiles(url, num_threads=num_threads, green=green)
   
@@ -59,6 +55,9 @@ def test_read_write(s3, protocol, num_threads, green):
   assert cf['info2', :] == content[:]
   assert cf.get('nonexistentfile') is None
 
+  assert cf.get('info', return_dict=True) == { "info": content }
+  assert cf.get(['info', 'info2'], return_dict=True) == { "info": content, "info2": content }
+
   del cf['info2']
   assert cf.exists('info2') == False
 
@@ -80,7 +79,23 @@ def test_read_write(s3, protocol, num_threads, green):
   cf.delete('info')
 
   if protocol == 'file':
-    rmtree(path)
+    rmtree(url)
+
+@pytest.mark.parametrize("protocol", ('mem', 'file', 's3'))#'gs'))
+def test_get_json_order(s3, protocol):
+  from cloudfiles import CloudFiles
+  url = compute_url(protocol, 'get_json_order')
+  cf = CloudFiles(url)
+
+  N = 5300
+  cf.put_jsons(( (str(z), [ z ]) for z in range(N) ))
+
+  contents = cf.get_json(( str(z) for z in range(N) ))
+
+  for z, content in enumerate(contents):
+    assert content[0] == z
+
+  cf.delete(( str(z) for z in range(N) ))
 
 @pytest.mark.parametrize("green", (False, True))
 @pytest.mark.parametrize("compress", (None, 'gzip','br','zstd'))
@@ -344,6 +359,20 @@ def test_exists(s3, protocol):
 
   cf.delete('info')
 
+@pytest.mark.parametrize("protocol", ('mem', 'file', 's3'))
+def test_isdir(s3, protocol):
+  from cloudfiles import CloudFiles, exceptions
+  url = compute_url(protocol, "isdir")
+
+  cf = CloudFiles(url, num_threads=5)
+  assert not cf.isdir()
+
+  content = b'some_string'
+  cf.put('info', content, compress=None)
+  
+  assert cf.isdir()
+  cf.delete('info')
+
 @pytest.mark.parametrize("protocol", ('mem', 'file', 's3'))
 def test_access_non_cannonical_paths(s3, protocol):
   from cloudfiles import CloudFiles, exceptions
@@ -478,6 +507,12 @@ def test_path_extraction():
   assert path.path == '/tmp/removeme/layer'
   assert path.host is None
 
+  assert (paths.extract('gs://username/a/username2/b/c/d') 
+      == ExtractedPath(
+        'precomputed', 'gs', 'username', 
+        'a/username2/b/c/d', None
+      ))
+
 @pytest.mark.parametrize("protocol", ('mem', 'file', 's3'))
 def test_access_non_cannonical_minimal_path(s3, protocol):
   from cloudfiles import CloudFiles, exceptions
@@ -650,6 +685,94 @@ def test_ascloudpath():
   assert ascloudpath(pth) == "precomputed://s3://https://some.domain.com/my_bucket/of/heaven"
 
 
+def test_cli_cp():
+  import subprocess
+  from cloudfiles.lib import mkdir, touch
+  test_dir = os.path.dirname(os.path.abspath(__file__))
+  srcdir = os.path.join(test_dir, "testfiles_src")
+  destdir = os.path.join(test_dir, "testfiles_dest")
+  N = 100
+
+  try:
+    shutil.rmtree(srcdir)
+  except FileNotFoundError:
+    pass
+
+  if os.path.isfile(f"{destdir}"):
+    os.remove(destdir)
+
+  try:
+    shutil.rmtree(destdir)
+  except FileNotFoundError:
+    pass
+
+  def mkfiles(mkdirname):
+    try:
+      shutil.rmtree(mkdirname)
+    except FileNotFoundError:
+      pass
+    mkdir(mkdirname)
+    for i in range(N):
+      touch(os.path.join(mkdirname, str(i)))
+
+  mkfiles(srcdir)
+
+  # cp -r has different behavior depending on if the dest 
+  # directory exists or not so we run the test twice
+  # in both conditions.
+  subprocess.run(["cloudfiles", "cp", "-r", srcdir, destdir])
+  assert len(os.listdir(srcdir)) == N
+  assert os.listdir(srcdir) == os.listdir(destdir)
+
+  subprocess.run(["cloudfiles", "cp", "-r", srcdir, destdir])
+  assert len(os.listdir(srcdir)) == N
+  assert os.listdir(srcdir) == os.listdir(os.path.join(destdir, os.path.basename(srcdir)))
+  shutil.rmtree(destdir)
+
+  subprocess.run(["cloudfiles", "cp", "-r", srcdir  + "/1*", destdir])
+  assert set(os.listdir(destdir)) == set([ "1"] + [ str(i) for i in range(10, 20) ])
+
+  shutil.rmtree(destdir)
+  subprocess.run(["cloudfiles", "cp", srcdir + "/**", destdir])
+  assert os.listdir(srcdir) == os.listdir(destdir)
+
+  shutil.rmtree(destdir)
+  subprocess.run(["cloudfiles", "cp", srcdir + "/*", destdir])
+  assert os.listdir(srcdir) == os.listdir(destdir)
+
+  shutil.rmtree(destdir)
+  subprocess.run(f"find {srcdir} -type f | cloudfiles cp - {destdir}", shell=True)
+  assert os.listdir(srcdir) == os.listdir(destdir)
+
+  shutil.rmtree(destdir)
+  mkdir(destdir)
+  subprocess.run(f"cloudfiles cp {srcdir}/10 {destdir}", shell=True)
+  assert os.listdir(destdir) == ["10"]
+
+  shutil.rmtree(destdir)
+  subprocess.run(f"cloudfiles cp {srcdir}/10 {destdir}", shell=True)
+  assert os.path.isfile(f"{destdir}")
+  os.remove(destdir)
+
+  mkdir(destdir)
+  subprocess.run(f"cloudfiles cp {srcdir}/8 {srcdir}/9 {destdir}", shell=True)
+  assert sorted(os.listdir(destdir)) == ["8","9"]
+  shutil.rmtree(destdir)
+
+  subprocess.run(f"cloudfiles cp {destdir}", shell=True)
+  assert not os.path.exists(destdir)
+
+  try:
+    shutil.rmtree(srcdir)
+  except FileNotFoundError:
+    pass
+
+  try:
+    shutil.rmtree(destdir)
+  except FileNotFoundError:
+    pass
+
+
 def test_cli_rm():
   import subprocess
   from cloudfiles.lib import mkdir, touch
@@ -712,3 +835,37 @@ def test_cli_rm():
   except FileNotFoundError:
     pass
 
+@pytest.mark.parametrize("green", (True, False))
+def test_exceptions_raised(green):
+  from cloudfiles import CloudFiles, exceptions
+  from cloudfiles.lib import mkdir
+  path = compute_url("file", "exceptions_raised")
+  cf = CloudFiles(path, green=green)
+
+  pth = mkdir(path.replace("file://", ""))
+  with open(f"{pth}/wontdecompress.gz", "wb") as f:
+    f.write(b"not a valid gzip stream")
+
+  try:
+    x = cf.get("wontdecompress")
+    assert False
+  except exceptions.DecompressionError:
+    pass
+
+  try:
+    x = cf.get(["wontdecompress"], raise_errors=True)
+    assert False
+  except exceptions.DecompressionError:
+    pass
+
+  try:
+    x = cf.get(["wontdecompress"], return_dict=True)
+    assert False
+  except exceptions.DecompressionError:
+    pass
+
+  cf.delete("wontdecompress")
+
+
+
+
