@@ -9,5 +9,5 @@ Amazon S3 interfaces, and reading from arbitrary HTTP
 servers.
 """
 
-from .cloudfiles import CloudFiles
+from .cloudfiles import CloudFiles, dl
 from .interfaces import reset_connection_pools
\ No newline at end of file
