@@ -1,13 +1,18 @@
 from queue import Queue
 from collections import defaultdict
+from functools import partial, wraps
+import inspect
+import math
+import multiprocessing
 import itertools
 import os.path
+import platform
 import posixpath
 import re
-from functools import partial
 import types
 
 import orjson
+import pathos.pools
 from tqdm import tqdm
 
 import google.cloud.storage 
@@ -38,6 +43,118 @@ INTERFACES = {
   'mem': MemoryInterface,
 }
 
+def parallelize(desc=None, returns_list=False):
+  """
+  @parallelize 
+
+  Allow multiprocess for methods that don't need to return anything.
+
+  desc: progress bar label
+  returns_list: accumulate results in a list and return it
+
+  Enables the "parallel" argument to do something.
+  """
+  def decor(fn):
+    @wraps(fn)
+    def inner_decor(*args, **kwargs):
+      nonlocal fn
+      nonlocal desc
+      nonlocal returns_list
+
+      while isinstance(fn, partial):
+        fn = fn.func
+
+      try:
+        sig = inspect.signature(fn).bind(*args, **kwargs)
+        parallel = sig.arguments.get("parallel", None)
+      except TypeError:
+        parallel = kwargs.get("parallel", None)
+        del kwargs["parallel"]
+        sig = inspect.signature(fn).bind_partial(*args, **kwargs)
+
+      params = sig.arguments
+      self = params.get("self", None)
+      if "self" in params:
+        del params["self"]
+      input_key, input_value = first(params.items())
+      del params[input_key]
+
+      parallel = nvl(parallel, self.parallel, 1)
+
+      if parallel == 1:
+        return fn(*args, **kwargs)
+
+      progress = params.get("progress", False)
+      params["progress"] = False
+      total = params.get("total", None)
+
+      if self is None:
+        fn = partial(fn, **params)
+      else:
+        fn = partial(fn, self, **params)
+      
+      return parallel_execute(
+        fn, input_value, parallel, total, progress, 
+        desc=desc, returns_list=returns_list
+      )
+  
+    return inner_decor
+  return decor
+
+def parallel_execute(
+    fn, inputs, parallel, 
+    total, progress, desc,
+    returns_list
+  ):
+  if parallel == 0:
+    parallel = multiprocessing.cpu_count()
+  elif parallel < 0:
+    raise ValueError(f"parallel must be >= 0. Got: {parallel}")
+
+  total = totalfn(inputs, total)
+
+  if total is not None and total < 0:
+    raise ValueError(f"total must be positive. Got: {total}")
+
+  block_size = 250
+  if total is not None:
+    block_size = max(20, int(math.ceil(total / parallel)))
+    block_size = max(block_size, 1)
+    block_size = min(1250, block_size)
+
+  if isinstance(progress, tqdm):
+    pbar = progress
+  else:
+    pbar = tqdm(desc=desc, total=total, disable=(not progress))
+
+  # Fix for MacOS which can segfault due to 
+  # urllib calling libdispatch which is not fork-safe
+  # https://bugs.python.org/issue30385
+  no_proxy = os.environ.get("no_proxy", "")
+  if platform.system().lower() == "darwin":
+    os.environ["no_proxy"] = "*"
+
+  results = []
+  try: 
+    with pathos.pools.ProcessPool(parallel) as executor:
+      for res in executor.imap(fn, sip(inputs, block_size)):
+        if isinstance(res, int):
+          pbar.update(res)
+        elif isinstance(res, list):
+          pbar.update(len(res))
+        else:
+          pbar.update(block_size)
+
+        if returns_list:
+          results.extend(res)
+  finally:  
+    if platform.system().lower() == "darwin":
+      os.environ["no_proxy"] = no_proxy
+    pbar.close()
+
+  if returns_list:
+    return results
+
 def get_interface_class(protocol):
   if protocol in INTERFACES:
     return INTERFACES[protocol]
@@ -59,6 +176,44 @@ def totalfn(files, total):
   except TypeError:
     return None
   
+def dl(cloudpaths, raw=False, **kwargs):
+  """
+  Shorthand for downloading files with CloudFiles.get.
+  You can use full paths from multiple buckets and services
+  and they'll be noted with key "fullpath" in each returned 
+  dict.
+  """
+  cloudpaths, is_multiple = toiter(cloudpaths, is_iter=True)
+  clustered = defaultdict(list)
+  total = 0
+  for path in cloudpaths:
+    epath = paths.extract(path)
+    bucketpath = paths.asbucketpath(epath)
+    clustered[bucketpath].append(epath.path)
+    total += 1
+
+  progress = kwargs.get("progress", False) and total > 1
+  pbar = tqdm(total=total, desc="Downloading", disable=(not progress))
+
+  try:
+    del kwargs["progress"]
+  except KeyError:
+    pass
+
+  content = []
+  for bucketpath, keys in clustered.items():
+    cf = CloudFiles(bucketpath, **kwargs)
+    sub_content = cf.get(keys, raw=raw, progress=pbar)
+    for sct in sub_content:
+      sct["fullpath"] = posixpath.join(bucketpath, sct["path"])
+    content += sub_content
+
+  pbar.close()
+
+  if not is_multiple:
+    return content[0]
+  return content
+
 class CloudFiles(object):
   """
   CloudFiles is a multithreaded key-value object
@@ -73,7 +228,8 @@ class CloudFiles(object):
   def __init__(
     self, cloudpath, progress=False, 
     green=False, secrets=None, num_threads=20,
-    use_https=False, endpoint=None
+    use_https=False, endpoint=None, parallel=1,
+    request_payer=None
   ):
     if use_https:
       cloudpath = paths.to_https_protocol(cloudpath)
@@ -83,6 +239,8 @@ class CloudFiles(object):
     self.secrets = secrets
     self.num_threads = num_threads
     self.green = bool(green)
+    self.parallel = int(parallel)
+    self.request_payer = request_payer
 
     self._path = paths.extract(cloudpath)
     if endpoint:
@@ -102,11 +260,23 @@ class CloudFiles(object):
     return self._interface_cls(
       self._path, 
       secrets=self.secrets,
+      request_payer=self.request_payer
     )
 
-  def get(self, paths, total=None, raw=False, progress=None):
+  def abspath(self, path):
+    sep = posixpath.sep
+    if self._path.protocol == "file":
+      sep = os.path.sep
+    return sep.join((paths.asprotocolpath(self._path), path))
+
+  @parallelize(desc="Download", returns_list=True)
+  def get(
+    self, paths, total=None, 
+    raw=False, progress=None, parallel=None,
+    return_dict=False, raise_errors=True
+  ):
     """
-    Download one or more files.
+    Download one or more files. Return order is not guaranteed to match input.
 
     paths: scalar or iterable of:
       filename (strings)
@@ -115,10 +285,17 @@ class CloudFiles(object):
     total: manually provide a progress bar size if paths does
       not support the `len` operator.
     raw: download without decompressing
-    
+    parallel: number of concurrent processes (0 means all cores)
+    return_dict: Turn output into { path: binary } and drops the 
+      extra information. Errors will be raised immediately.
+    raise_errors: Raise the first error immediately instead 
+      of returning them as part of the output.
+
     Returns:
-      if paths is scalar:
-        binary
+      if return_dict:
+        { path: binary, path: binary } (errors raised immediately)
+      elif paths is scalar:
+        binary (errors raised immediately)
       else:
         [
           {
@@ -133,6 +310,9 @@ class CloudFiles(object):
     """
     paths, multiple_return = toiter(paths, is_iter=True)
     progress = nvl(progress, self.progress)
+    # return_dict prevents the user from having a chance
+    # to inspect errors, so we must raise here.
+    raise_errors = raise_errors or return_dict or (not multiple_return)
 
     def check_md5(path, content, server_hash):
       if server_hash is None:
@@ -178,6 +358,9 @@ class CloudFiles(object):
       except Exception as err:
         error = err
 
+      if raise_errors and error:
+        raise error
+
       return { 
         'path': path, 
         'content': content, 
@@ -190,15 +373,15 @@ class CloudFiles(object):
     total = totalfn(paths, total)
 
     if total == 1:
-      ret = download(paths[0])
-      if multiple_return:
+      ret = download(first(paths))
+      if return_dict:
+        return { ret["path"]: ret["content"] }
+      elif multiple_return:
         return [ ret ]
-      elif ret['error']:
-        raise ret['error']
       else:
         return ret['content']
 
-    return schedule_jobs(
+    results = schedule_jobs(
       fns=( partial(download, path) for path in paths ), 
       concurrency=self.num_threads, 
       progress=progress,
@@ -206,9 +389,15 @@ class CloudFiles(object):
       green=self.green,
     )
 
+    if return_dict:
+      return { res["path"]: res["content"] for res in results }  
+
+    return results
+
   def get_json(self, paths, total=None):
     """
     Download one or more JSON files and decode them into a python object.
+    Return order is guaranteed to match input.
 
     paths: scalar or iterable of:
       filename (strings)
@@ -220,19 +409,11 @@ class CloudFiles(object):
 
     Returns:
       if paths is scalar:
-        object
+        decoded json object
       else:
-        [
-          {
-            'path': path, 
-            'content': object, 
-            'byte_range': (start, end),
-            'error': error,
-          }
-        ]
+        [ decoded json objects ]
     """
     paths, multiple_return = toiter(paths, is_iter=True)
-    contents = self.get(paths, total=total)
 
     def decode(content):
       content = content['content']
@@ -240,19 +421,27 @@ class CloudFiles(object):
         return None
       return orjson.loads(content.decode('utf8'))
 
-    if not multiple_return and contents and contents[0]['error']:
-      raise contents[0]['error']
+    desc = self.progress if isinstance(self.progress, str) else "Downloading JSON"
+
+    contents = []
+    with tqdm(total=totalfn(paths, total), desc=desc, disable=(not self.progress)) as pbar:
+      for paths_chunk in sip(paths, 2000):
+        contents_chunk = self.get(paths_chunk, total=total, progress=pbar)
+        pathidx = { content["path"]: content for content in contents_chunk }
+        contents.extend(( pathidx[pth] for pth in paths_chunk ))
 
     contents = [ decode(content) for content in contents ]
     if multiple_return:
       return contents
     return contents[0]
 
+  @parallelize(desc="Upload")
   def puts(
     self, files, 
     content_type=None, compress=None, 
     compression_level=None, cache_control=None,
-    total=None, raw=False, progress=None
+    total=None, raw=False, progress=None,
+    parallel=1, storage_class=None
   ):
     """
     Writes one or more files at a given location.
@@ -270,6 +459,8 @@ class CloudFiles(object):
           'compression_level': e.g. 6, # for gzip, brotli, or zstd
           'cache_control': specify the header the way you want 
               e.g. 'no-cache' or 'public; max-age=3600' etc
+          'storage_class': for relevant cloud providers, specify the storage class of the file
+              e.g. 'STANDARD', 'COLDLINE', etc. (note: these are vendor-specific)
         }
 
       If the additional fields are specified, they will override the 
@@ -285,6 +476,7 @@ class CloudFiles(object):
     progress: selectively enable or disable progress just for this
       function call. If progress is a string, it sets the 
       text of the progress bar.
+    parallel: number of concurrent processes (0 means all cores)
     """
     files = toiter(files)
     progress = nvl(progress, self.progress)
@@ -316,6 +508,7 @@ class CloudFiles(object):
           content_type=file.get('content_type', content_type),
           compress=file_compress,
           cache_control=file.get('cache_control', cache_control),
+          storage_class=file.get('storage_class', storage_class)
         )
 
     if not isinstance(files, types.GeneratorType):
@@ -330,21 +523,22 @@ class CloudFiles(object):
       return
 
     fns = ( partial(uploadfn, file) for file in files )
-    desc = self._progress_description('Uploading')
-    schedule_jobs(
+    desc = self._progress_description("Upload")
+    results = schedule_jobs(
       fns=fns,
       concurrency=self.num_threads,
       progress=(desc if progress else None),
       total=total,
       green=self.green,
     )
+    return len(results)
 
   def put(
     self, 
     path, content,     
     content_type=None, compress=None, 
     compression_level=None, cache_control=None,
-    raw=False
+    raw=False, storage_class=None
   ):
     """
     Write a single file.
@@ -355,6 +549,7 @@ class CloudFiles(object):
     compress: None, 'gzip', 'br' (brotli), 'zstd'
     compression_level: (None or int) input to compressor, None means use default
     cache_control: (str) HTTP Cache-Control header.
+    storage_class: (str) Storage class for the file.
     raw: (bool) if true, content is pre-compressed and 
       will bypass the compressor
 
@@ -367,13 +562,15 @@ class CloudFiles(object):
       'compress': compress,
       'compression_level': compression_level,
       'cache_control': cache_control,
+      'storage_class': storage_class,
     }, raw=raw)
 
   def put_jsons(
     self, files,     
     compress=None, compression_level=None, 
-    cache_control=None, total=None, raw=False,
-    progress=None
+    cache_control=None, total=None,
+    raw=False, progress=None, parallel=1,
+    storage_class=None
   ):
     """
     Write one or more files as JSON.
@@ -400,14 +597,15 @@ class CloudFiles(object):
     return self.puts( 
       (jsonify_file(file) for file in files), 
       compress=compress, compression_level=compression_level,
-      content_type='application/json', total=total, raw=raw,
-      progress=progress
+      content_type='application/json', storage_class=storage_class,
+      total=total, raw=raw,
+      progress=progress, parallel=parallel
     )
 
   def put_json(
     self, path, content,
     compress=None, compression_level=None, 
-    cache_control=None
+    cache_control=None, storage_class=None
   ):
     """
     Write a single JSON file. Automatically supplies the
@@ -418,6 +616,7 @@ class CloudFiles(object):
     compress: None, 'gzip', 'br' (brotli), 'zstd'
     compression_level: (None or int) input to compressor, None means use default
     cache_control: (str) HTTP Cache-Control header.
+    storage_class: (str) Storage class for the file.
 
     Returns: void
     """
@@ -428,8 +627,25 @@ class CloudFiles(object):
       'compress': compress,
       'compression_level': compression_level,
       'cache_control': cache_control,
+      'storage_class': storage_class
     })
 
+  def isdir(self, prefix=""):
+    """
+    Tests if the given path points to a directory.
+    This has a typical meaning on a filesystem,
+    but on object storage, since directories don't
+    exist, it just means "is there at least one object
+    located at or below this path".
+    """
+    if self._path.protocol == "file":
+      fullpath = paths.asfilepath(self._path)
+      fullpath = os.path.join(fullpath, prefix)
+      return os.path.isdir(fullpath)
+
+    res = first(self.list(prefix=prefix))
+    return res is not None
+
   def exists(self, paths, total=None, progress=None):
     """
     Test if the given file paths exist.
@@ -455,15 +671,43 @@ class CloudFiles(object):
     progress = nvl(progress, self.progress)
     results = {}
 
+    total = totalfn(paths, total)
+    desc = self._progress_description("Exists")
+    if isinstance(progress, tqdm):
+      pbar = progress
+    else:
+      pbar = tqdm(total=total, desc=desc, disable=(not progress))
+
     def exist_thunk(paths):
       with self._get_connection() as conn:
         results.update(conn.files_exist(paths))
+        pbar.update(len(paths))
 
     batch_size = self._interface_cls.exists_batch_size
-    
-    desc = self._progress_description('Existence Testing')
     schedule_jobs(
       fns=( partial(exist_thunk, paths) for paths in sip(paths, batch_size) ),
+      progress=False,
+      concurrency=self.num_threads,
+      green=self.green,
+    )
+    pbar.close()
+
+    if return_multiple:
+      return results
+    return first(results.values())
+
+  def head(self, paths, total=None, progress=None):
+    paths, return_multiple = toiter(paths, is_iter=True)
+    progress = nvl(progress, self.progress)
+    results = {}
+
+    def size_thunk(path):
+      with self._get_connection() as conn:
+        results[path] = conn.head(path)
+    
+    desc = self._progress_description('Head')
+    schedule_jobs(
+      fns=( partial(size_thunk, path) for path in paths ),
       progress=(desc if progress else None),
       concurrency=self.num_threads,
       total=totalfn(paths, total),
@@ -472,7 +716,7 @@ class CloudFiles(object):
 
     if return_multiple:
       return results
-    return first(results.values())
+    return first(results.values())   
 
   def size(self, paths, total=None, progress=None):
     """
@@ -499,7 +743,8 @@ class CloudFiles(object):
       return results
     return first(results.values())
 
-  def delete(self, paths, total=None, progress=None):
+  @parallelize(desc="Delete")
+  def delete(self, paths, total=None, progress=None, parallel=1):
     """
     Delete one or more files.
 
@@ -510,6 +755,7 @@ class CloudFiles(object):
     progress: selectively enable or disable progress just for this
       function call. If progress is a string, it sets the 
       text of the progress bar.
+    parallel: number of concurrent processes (0 means all cores)
 
     Returns: void
     """
@@ -520,17 +766,18 @@ class CloudFiles(object):
       with self._get_connection() as conn:
         conn.delete_files(path)
 
-    desc = self._progress_description('Deleting')
+    desc = self._progress_description('Delete')
 
     batch_size = self._interface_cls.delete_batch_size
 
-    schedule_jobs(
+    results = schedule_jobs(
       fns=( partial(thunk_delete, path) for path in sip(paths, batch_size) ),
       progress=(desc if progress else None),
       concurrency=self.num_threads,
       total=totalfn(paths, total),
       green=self.green,
     )
+    return len(results)
 
   def list(self, prefix="", flat=False):
     """
@@ -612,6 +859,15 @@ class CloudFiles(object):
         downloaded = cf_src.get(block_paths, raw=True, progress=False)
         if reencode is not None:
           downloaded = compression.transcode(downloaded, reencode, in_place=True)
+
+        if (
+              cf_src._path.protocol == "file"
+          and self._path.protocol != "file"
+          and os.path.sep != posixpath.sep
+        ):
+          for download in downloaded:
+            download["path"] = posixpath.sep.join(download["path"].split(os.path.sep))
+
         self.puts(downloaded, raw=True, progress=False)
         pbar.update(len(block_paths))
 
@@ -640,4 +896,3 @@ class CloudFiles(object):
 
   def __iter__(self):
     return self.list()
-
