@@ -1,6 +1,7 @@
 import base64
 import binascii
-from collections import defaultdict
+from collections import defaultdict, namedtuple
+from datetime import datetime
 import json
 import os.path
 import posixpath
@@ -16,7 +17,8 @@ import tenacity
 
 from .compression import COMPRESSION_TYPES
 from .connectionpools import S3ConnectionPool, GCloudBucketPool, MemoryPool, MEMORY_DATA
-from .lib import mkdir, sip, md5, PYTHON3
+from .exceptions import MD5IntegrityError
+from .lib import mkdir, sip, md5, PYTHON3, validate_s3_multipart_etag
 
 COMPRESSION_EXTENSIONS = ('.gz', '.br', '.zstd')
 GZIP_TYPES = (True, 'gzip', 1)
@@ -33,14 +35,21 @@ class keydefaultdict(defaultdict):
 S3_POOL = None
 GC_POOL = None
 MEM_POOL = None
+
+S3ConnectionPoolParams = namedtuple('S3ConnectionPoolParams', 'service bucket_name request_payer')
+GCloudBucketPoolParams = namedtuple('GCloudBucketPoolParams', 'bucket_name request_payer')
+MemoryPoolParams = namedtuple('MemoryPoolParams', 'bucket_name')
+
 def reset_connection_pools():
   global S3_POOL
   global GC_POOL
   global MEM_POOL
-  S3_POOL = keydefaultdict(lambda service: keydefaultdict(lambda bucket_name: S3ConnectionPool(service, bucket_name)))
-  GC_POOL = keydefaultdict(lambda bucket_name: GCloudBucketPool(bucket_name))
-  MEM_POOL = keydefaultdict(lambda bucket_name: MemoryPool(bucket_name))
+  S3_POOL = keydefaultdict(lambda params: S3ConnectionPool(params.service, params.bucket_name))
+  GC_POOL = keydefaultdict(lambda params: GCloudBucketPool(params.bucket_name, params.request_payer))
+  MEM_POOL = keydefaultdict(lambda params: MemoryPool(params.bucket_name))
   MEMORY_DATA = {}
+  import gc
+  gc.collect()
 
 reset_connection_pools()
 
@@ -61,9 +70,11 @@ class StorageInterface(object):
     self.release_connection()
 
 class FileInterface(StorageInterface):
-  def __init__(self, path, secrets=None):
+  def __init__(self, path, secrets=None, request_payer=None):
     super(StorageInterface, self).__init__()
     self._path = path
+    if request_payer is not None:
+      raise ValueError("Specifying a request payer for the FileInterface is not supported. request_payer must be None, got '{}'.".format(request_payer))
 
   def get_path_to_file(self, file_path):
     return os.path.join(self._path.path, file_path)
@@ -71,7 +82,8 @@ class FileInterface(StorageInterface):
   def put_file(
     self, file_path, content, 
     content_type, compress, 
-    cache_control=None
+    cache_control=None,
+    storage_class=None
   ):
     path = self.get_path_to_file(file_path)
     mkdir(os.path.dirname(path))
@@ -100,6 +112,41 @@ class FileInterface(StorageInterface):
       with open(path, 'wb') as f:
         f.write(content)
 
+  def head(self, file_path):
+    path = self.get_path_to_file(file_path)
+
+    if os.path.exists(path + '.gz'):
+      encoding = "gzip"
+      path += '.gz'
+    elif os.path.exists(path + '.br'):
+      encoding = "br"
+      path += ".br"
+    elif os.path.exists(path + '.zstd'):
+      encoding = "zstd"
+      path += ".zstd"
+    else:
+      encoding = None   
+
+    try:
+      statinfo = os.stat(path)
+    except FileNotFoundError:
+      return None
+
+    return {
+      "Cache-Control": None,
+      "Content-Length": statinfo.st_size,
+      "Content-Type": None,
+      "ETag": None,
+      "Last-Modified": datetime.utcfromtimestamp(statinfo.st_mtime),
+      "Content-Md5": None,
+      "Content-Encoding": encoding,
+      "Content-Disposition": None,
+      "Content-Language": None,
+      "Storage-Class": None,
+      "Request-Charged": None,
+      "Parts-Count": None,
+    }
+
   def get_file(self, file_path, start=None, end=None):
     path = self.get_path_to_file(file_path)
 
@@ -182,8 +229,8 @@ class FileInterface(StorageInterface):
     filenames = []
 
     remove = layer_path
-    if len(remove) and remove[-1] != '/':
-      remove += '/'
+    if len(remove) and remove[-1] != os.path.sep:
+      remove += os.path.sep
 
     if flat:
       for file_path in glob(path):
@@ -212,20 +259,23 @@ class FileInterface(StorageInterface):
     return iter(_radix_sort(filenames))
 
 class MemoryInterface(StorageInterface):
-  def __init__(self, path, secrets=None):
+  def __init__(self, path, secrets=None, request_payer=None):
     super(StorageInterface, self).__init__()
+    if request_payer is not None:
+      raise ValueError("Specifying a request payer for the MemoryInterface is not supported. request_payer must be None, got '{}'.", request_payer)
     self._path = path
-    self._data = MEM_POOL[path.bucket].get_connection(secrets, None)
+    self._data = MEM_POOL[MemoryPoolParams(path.bucket)].get_connection(secrets, None)
 
   def get_path_to_file(self, file_path):
-    return os.path.join(
+    return posixpath.join(
       self._path.path, file_path
     )
 
   def put_file(
     self, file_path, content, 
     content_type, compress, 
-    cache_control=None
+    cache_control=None,
+    storage_class=None
   ):
     path = self.get_path_to_file(file_path)
 
@@ -268,6 +318,9 @@ class MemoryInterface(StorageInterface):
       result = result[slice(start, end)]
     return (result, encoding, None, None)
 
+  def head(self, file_path):
+    raise NotImplementedError()
+
   def size(self, file_path):
     path = self.get_path_to_file(file_path)
 
@@ -348,17 +401,19 @@ class GoogleCloudStorageInterface(StorageInterface):
   exists_batch_size = Batch._MAX_BATCH_SIZE
   delete_batch_size = Batch._MAX_BATCH_SIZE
 
-  def __init__(self, path, secrets=None):
+  def __init__(self, path, secrets=None, request_payer=None):
     super(StorageInterface, self).__init__()
     global GC_POOL
     self._path = path
-    self._bucket = GC_POOL[path.bucket].get_connection(secrets, None)
+    self._request_payer = request_payer
+    self._bucket = GC_POOL[GCloudBucketPoolParams(self._path.bucket, self._request_payer)].get_connection(secrets, None)
     
   def get_path_to_file(self, file_path):
     return posixpath.join(self._path.path, file_path)
 
   @retry
-  def put_file(self, file_path, content, content_type, compress, cache_control=None):
+  def put_file(self, file_path, content, content_type,
+               compress, cache_control=None, storage_class=None):
     key = self.get_path_to_file(file_path)
     blob = self._bucket.blob( key )
 
@@ -366,11 +421,15 @@ class GoogleCloudStorageInterface(StorageInterface):
       blob.content_encoding = "br"
     elif compress in GZIP_TYPES:
       blob.content_encoding = "gzip"
+    elif compress == "zstd":
+      blob.content_encoding = "zstd"
     elif compress:
       raise ValueError("Compression type {} not supported.".format(compress))
 
     if cache_control:
       blob.cache_control = cache_control
+    if storage_class:
+      blob.storage_class = storage_class
 
     blob.md5_hash = md5(content)
     blob.upload_from_string(content, content_type)
@@ -386,7 +445,7 @@ class GoogleCloudStorageInterface(StorageInterface):
       end = int(end - 1)
 
     try:
-      content = blob.download_as_string(start=start, end=end, raw_download=True)
+      content = blob.download_as_bytes(start=start, end=end, raw_download=True)
     except google.cloud.exceptions.NotFound as err:
       return (None, None, None, None)
 
@@ -399,6 +458,25 @@ class GoogleCloudStorageInterface(StorageInterface):
 
     return (content, blob.content_encoding, hash_value, hash_type)
 
+  @retry
+  def head(self, file_path):
+    key = self.get_path_to_file(file_path)
+    blob = self._bucket.get_blob(key)
+    return {
+      "Cache-Control": blob.cache_control,
+      "Content-Length": blob.size,
+      "Content-Type": blob.content_type,
+      "ETag": blob.etag,
+      "Last-Modified": blob.time_created,
+      "Content-Md5": blob.md5_hash,
+      "Content-Crc32c": blob.crc32c,
+      "Content-Encoding": blob.content_encoding,
+      "Content-Disposition": blob.content_disposition,
+      "Content-Language": blob.content_language,
+      "Storage-Class": blob.storage_class,
+      "Component-Count": blob.component_count,
+    }
+
   @retry
   def size(self, file_path):
     key = self.get_path_to_file(file_path)
@@ -462,7 +540,9 @@ class GoogleCloudStorageInterface(StorageInterface):
     """
     layer_path = self.get_path_to_file("")        
     path = posixpath.join(layer_path, prefix)
-    for blob in self._bucket.list_blobs(prefix=path):
+
+    delimiter = '/' if flat else None
+    for blob in self._bucket.list_blobs(prefix=path, delimiter=delimiter):
       filename = blob.name.replace(layer_path, '')
       if not filename:
         continue
@@ -473,12 +553,14 @@ class GoogleCloudStorageInterface(StorageInterface):
 
   def release_connection(self):
     global GC_POOL
-    GC_POOL[self._path.bucket].release_connection(self._bucket)
+    GC_POOL[GCloudBucketPoolParams(self._path.bucket, self._request_payer)].release_connection(self._bucket)
 
 class HttpInterface(StorageInterface):
-  def __init__(self, path, secrets=None):
+  def __init__(self, path, secrets=None, request_payer=None):
     super(StorageInterface, self).__init__()
     self._path = path
+    if request_payer is not None:
+      raise ValueError("Specifying a request payer for the HttpInterface is not supported. request_payer must be None, got '{}'.".format(request_payer))
 
   def get_path_to_file(self, file_path):
     return posixpath.join(self._path.host, self._path.path, file_path)
@@ -491,9 +573,17 @@ class HttpInterface(StorageInterface):
     raise NotImplementedError()
 
   # @retry
-  def put_file(self, file_path, content, content_type, compress, cache_control=None):
+  def put_file(self, file_path, content, content_type,
+               compress, cache_control=None, storage_class=None):
     raise NotImplementedError()
 
+  @retry
+  def head(self, file_path):
+    key = self.get_path_to_file(file_path)
+    resp = requests.head(key)
+    resp.raise_for_status()
+    return resp.headers
+
   @retry
   def get_file(self, file_path, start=None, end=None):
     key = self.get_path_to_file(file_path)
@@ -538,11 +628,20 @@ class S3Interface(StorageInterface):
   # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Bucket.delete_objects
   # claims batch size limit is 1000
   delete_batch_size = 1000
-  def __init__(self, path, secrets=None):
+  def __init__(self, path, secrets=None, request_payer=None):
     super(StorageInterface, self).__init__()
     global S3_POOL
+
+    if request_payer is None:
+      self._additional_attrs = {}
+    elif request_payer == 'requester':
+      self._additional_attrs = {'RequestPayer': request_payer}
+    else:
+      raise ValueError("request_payer for S3Interface must either be None or 'requester', got '{}'.".format(request_payer))
+
+    self._request_payer = request_payer
     self._path = path
-    self._conn = S3_POOL[path.protocol][path.bucket].get_connection(secrets, path.host)
+    self._conn = S3_POOL[S3ConnectionPoolParams(path.protocol, path.bucket, request_payer)].get_connection(secrets, path.host)
 
   def get_path_to_file(self, file_path):
     return posixpath.join(self._path.path, file_path)
@@ -551,7 +650,9 @@ class S3Interface(StorageInterface):
   def put_file(
     self, file_path, content, 
     content_type, compress, 
-    cache_control=None, ACL="bucket-owner-full-control"
+    cache_control=None,
+    ACL="bucket-owner-full-control",
+    storage_class=None
   ):
     key = self.get_path_to_file(file_path)
 
@@ -562,6 +663,7 @@ class S3Interface(StorageInterface):
       'ContentType': (content_type or 'application/octet-stream'),
       'ACL': ACL,
       'ContentMD5': md5(content),
+      **self._additional_attrs,
     }
 
     # keep gzip as default
@@ -576,6 +678,8 @@ class S3Interface(StorageInterface):
 
     if cache_control:
       attrs['CacheControl'] = cache_control
+    if storage_class:
+      attrs['StorageClass'] = storage_class
 
     self._conn.put_object(**attrs)
 
@@ -587,8 +691,9 @@ class S3Interface(StorageInterface):
     None when the file doesn't exist.
     """
 
-    kwargs = {}
-    if start is not None or end is not None:
+    kwargs = self._additional_attrs.copy()
+    range_request = start is not None or end is not None
+    if range_request:
       start = int(start) if start is not None else ''
       end = int(end - 1) if end is not None else ''
       kwargs['Range'] = "bytes={}-{}".format(start, end)
@@ -607,24 +712,64 @@ class S3Interface(StorageInterface):
       # s3 etags return hex digests but we need the base64 encoding
       # to make uniform comparisons. 
       # example s3 etag: "31ee76261d87fed8cb9d4c465c48158c"
-      etag = resp.get('ETag', None)
-      if etag is not None:
-        etag = etag.lstrip('"').rstrip('"')
-        etag = base64.b64encode(binascii.unhexlify(etag)).decode('utf8')
+      # example multipart s3 etag: "cd8d2616dfa6cc80a06a846d3b3f6f30-14"
+      # The -14 means 14 parts.
 
-      return (resp['Body'].read(), encoding, etag, "md5")
+      etag = resp.get('ETag', None)
+      content = resp['Body'].read()
+
+      if etag is not None and not range_request:
+        etag = etag.lstrip('"').rstrip('"')
+        # AWS has special multipart validation
+        # so we handle it here... leaky abstraction.
+        if '-' in etag: 
+          if not validate_s3_multipart_etag(content, etag):
+            raise MD5IntegrityError(f"{file_path} failed its multipart md5 check. server md5: {etag}")
+          etag = None
+        else:
+          etag = base64.b64encode(binascii.unhexlify(etag)).decode('utf8')
+
+      return (content, encoding, etag, "md5")
     except botocore.exceptions.ClientError as err: 
       if err.response['Error']['Code'] == 'NoSuchKey':
         return (None, None, None, None)
       else:
         raise
 
+  @retry
+  def head(self, file_path):
+    try:
+      response = self._conn.head_object(
+        Bucket=self._path.bucket,
+        Key=self.get_path_to_file(file_path),
+        **self._additional_attrs,
+      )
+      return {
+        "Cache-Control": response.get("CacheControl", None),
+        "Content-Length": response.get("ContentLength", None),
+        "Content-Type": response.get("ContentType", None),
+        "ETag": response.get("ETag", None),
+        "Last-Modified": response.get("LastModified", None),
+        "Content-Md5": response["ResponseMetadata"]["HTTPHeaders"].get("content-md5", None),
+        "Content-Encoding": response.get("ContentEncoding", None),
+        "Content-Disposition": response.get("ContentDisposition", None),
+        "Content-Language": response.get("ContentLanguage", None),
+        "Storage-Class": response.get("StorageClass", None),
+        "Request-Charged": response.get("RequestCharged", None),
+        "Parts-Count": response.get("PartsCount", None),
+      }
+    except botocore.exceptions.ClientError as e:
+      if e.response['Error']['Code'] == "404":
+        return None
+      raise
+
   @retry
   def size(self, file_path):
     try:
       response = self._conn.head_object(
         Bucket=self._path.bucket,
         Key=self.get_path_to_file(file_path),
+        **self._additional_attrs,
       )
       return response['ContentLength']
     except botocore.exceptions.ClientError as e:
@@ -638,6 +783,7 @@ class S3Interface(StorageInterface):
       self._conn.head_object(
         Bucket=self._path.bucket,
         Key=self.get_path_to_file(file_path),
+        **self._additional_attrs,
       )
     except botocore.exceptions.ClientError as e:
       if e.response['Error']['Code'] == "404":
@@ -666,6 +812,7 @@ class S3Interface(StorageInterface):
     self._conn.delete_object(
       Bucket=self._path.bucket,
       Key=self.get_path_to_file(file_path),
+      **self._additional_attrs,
     )
 
   @retry
@@ -679,6 +826,7 @@ class S3Interface(StorageInterface):
           ],
           'Quiet': True
         },
+        **self._additional_attrs,
       )
 
   def list_files(self, prefix, flat=False):
@@ -698,7 +846,10 @@ class S3Interface(StorageInterface):
       kwargs = {
         'Bucket': self._path.bucket,
         'Prefix': path,
+        **self._additional_attrs
       }
+      if flat:
+        kwargs['Delimiter'] = '/'
 
       if continuation_token:
         kwargs['ContinuationToken'] = continuation_token
@@ -719,7 +870,6 @@ class S3Interface(StorageInterface):
         elif flat and '/' not in key.replace(path, ''):
           yield filename
 
-
     for filename in iterate(resp):
       yield filename
 
@@ -731,7 +881,7 @@ class S3Interface(StorageInterface):
 
   def release_connection(self):
     global S3_POOL
-    S3_POOL[self._path.protocol][self._path.bucket].release_connection(self._conn)
+    S3_POOL[S3ConnectionPoolParams(self._path.protocol, self._path.bucket, self._request_payer)].release_connection(self._conn)
 
 
 def _radix_sort(L, i=0):
